/*
 * Copyright (C) 2007 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include<iostream>
#include "ChoicePoint.hh"

using namespace std;

/**
 * constructor: make a copy of wb and affect it to worldbase, affect copies all other arguments to
 * their corresponding fields, sets tag to UNEXPLORED, sets father to NULL,
 * sets task to NULL
 */
ChoicePoint::ChoicePoint(string const& sType, LightWorldBase const& wb, TaskList const& TL,
	std::vector<unsigned long int> const& po, PlanTree const& PT, PartialPlanAppraiser const& pEval)
	: type(sType), tag(UNEXPLORED_TAG), worldbase(new LightWorldBase(wb)), tasklist(new TaskList(TL)), projection_order(po),
	task(EMPTY_ID), tree(new PlanTree(PT)), partial_eval(new PartialPlanAppraiser(pEval)), father(NULL)
{}

/**
 * constructor: make a copy of wb and affect it to worldbase, affect copies all other arguments to
 * their corresponding fields, sets tag to UNEXPLORED, sets father to NULL,
 * sets task to NULL
 */
ChoicePoint::ChoicePoint(string const& sType, WorldBase const& wb, TaskList const& TL,
	std::vector<unsigned long int> const& po, PlanTree const& PT, PartialPlanAppraiser const& pEval)
	: type(sType), tag(UNEXPLORED_TAG), worldbase(new LightWorldBase(wb)), tasklist(new TaskList(TL)), projection_order(po),
	task(EMPTY_ID), tree(new PlanTree(PT)), partial_eval(new PartialPlanAppraiser(pEval)), father(NULL)
{}

/**
 * constructor: make a copy of wb and affect it to worldbase, affect all other arguments to
 * their corresponding fields, sets tag to UNEXPLORED, sets father to NULL,
 * sets task to NULL
 */
ChoicePoint::ChoicePoint(string const& sType, LightWorldBase const& wb, TaskList* TL,
	std::vector<unsigned long int> const& po, PlanTree* PT, PartialPlanAppraiser const& pEval)
	: type(sType), tag(UNEXPLORED_TAG), worldbase(new LightWorldBase(wb)), tasklist(TL), projection_order(po),
	task(EMPTY_ID), tree(PT), partial_eval(new PartialPlanAppraiser(pEval)), father(NULL)
{}

/**
 * constructor: make a copy of wb and affect it to worldbase, affect copies of
 * all other arguments to * their corresponding fields, sets tag to UNEXPLORED, sets father to NULL,
 * sets task to corresponding member (with ID = TaskID) in the copy of
 * TL (constructor used for parrallelChoicePoint)
 */
ChoicePoint::ChoicePoint(string const& sType, LightWorldBase const& wb, TaskList const& TL,
	std::vector<unsigned long int> const& po, PlanTree const& PT, unsigned long int TaskID, PartialPlanAppraiser const& pEval)
	: type(sType), tag(UNEXPLORED_TAG), worldbase(new LightWorldBase(wb)), tasklist(new TaskList(TL)), projection_order(po),
	task(TaskID), tree(new PlanTree(PT)), partial_eval(new PartialPlanAppraiser(pEval)), father(NULL)
{
	if(sType != PARALLEL_TYPE) {
		cerr << "ERROR: trying to create a parallelChoicePoint with specified type: " << sType;
		cerr << " provided to the constructor (no warranty about the result)" << endl;
	}
}

/**
 * constructor: make a copy of wb and affect it to worldbase, affect all other arguments to
 * their corresponding fields, sets tag to UNEXPLORED, sets father to NULL,
 * sets task to corresponding member (with ID = TaskID) in the copy of
 * TL (constructor used for parrallelChoicePoint)
 */
ChoicePoint::ChoicePoint(string const& sType, LightWorldBase const& wb, TaskList* TL,
	std::vector<unsigned long int> const& po, PlanTree* PT, unsigned long int TaskID, PartialPlanAppraiser const& pEval)
	: type(sType), tag(UNEXPLORED_TAG), worldbase(new LightWorldBase(wb)), tasklist(TL), projection_order(po),
	task(TaskID), tree(PT), partial_eval(new PartialPlanAppraiser(pEval)), father(NULL)
{
	if(sType != PARALLEL_TYPE) {
		cerr << "ERROR: trying to create a parallelChoicePoint with specified type: " << sType;
		cerr << " provided to the constructor (no warranty about the result)" << endl;
	}
}

/**
 * constructor by copying
 */
ChoicePoint::ChoicePoint(ChoicePoint const& cp)
	: type(cp.getType()), tag(cp.getTag()), worldbase(new LightWorldBase(*(cp.getWorldBase()))),
	tasklist(new TaskList(*(cp.getTaskList()), false)), projection_order(cp.getProjectionOrder()),
	task(cp.getTask()), tree(new PlanTree(*(cp.getTree()))), partial_eval(new PartialPlanAppraiser(*(cp.getPartialEval())))
{}// fields father and next will be updated by the ChoiceReminder

/**
 * propagates tag from next to the current node
 * if all next are failed tag: tag = failed tag
 * if all next are success tag: tag = success tag
 * if all next are cost_exceeded tag: tag = cost_exceeded tag
 * if all next are success, failed or explored tag: tag = explored tag
 * then return tag;
 */
string ChoicePoint::propagateTags() {
	vector<string> vecTags(next.size());
	bool tag_updated = false;
	unsigned int i;
	for(i=0 ; i<next.size() ; ++i) vecTags[i] = next[i]->getTag();
	if(!tag_updated) { // propagation of EXPLORED_TAG
		for(i=0 ; i<vecTags.size() ; ++i)
			if(vecTags[i] != SUCCESS_TAG && vecTags[i] != FAILED_TAG && vecTags[i] != EXPLORED_TAG && vecTags[i] != COST_EXCEEDED_TAG)
				break;
		if(vecTags.size()>0 && i==vecTags.size()) {
			tag = EXPLORED_TAG;
			tag_updated = true;
		}
	}
	return tag;
}

/**
 * returns true if the current point has no child with tag CURRENT
 */
bool ChoicePoint::hasNoCurrentChild() const
{
	for(unsigned int i=0 ; i<next.size() ; ++i)
		if(next[i]->getTag() == CURRENT_TAG) return false;
	return true;
}
