#if !defined HATP_LIGHT_ATTRIBUTE_SET_HH_
#define HATP_LIGHT_ATTRIBUTE_SET_HH_

#include<string>
#include<list>
#include<set>
#include<vector>
#include<map>
#include<sstream>

#include "HATPexception.hh"
#include "AttributeSet.hh"

/** Represent all the attributes of the type T.
 * The attributes can be static or dynamic and atom or a set.
 */
template <typename T> class LightAttributeSet {
private:

	/** Default value for atom attributes.
	*/
	T defaultAtomValue;

	/** Default value for set attributes.
	*/
	std::set<T> defaultSetValue;

	/** The static atom attributes. It is a pointer that will point toward the AttributeSet.
	*/
	std::map<std::string, T>* staticAtomAttrs;

	/** The dynamic atom attributes.
	*/
	std::map<std::string, T> dynamicAtomAttrs;

	/** The static set attributes. It is a pointer that will point toward the AttributeSet.
	*/
	std::map<std::string, std::set<T> >* staticSetAttrs;

	/** The dynamic set attributes.
	*/
	std::map<std::string, std::set<T> > dynamicSetAttrs;

public:
	LightAttributeSet(T argDefaultValue);
	explicit LightAttributeSet(T argDefaultValue, std::map<std::string, T>* argStaticAtomAttrs,
		std::vector<std::string> const& argDynamicAtomAttrs,
		std::map<std::string, std::set<T> >* argStaticSetAttrs,
		std::vector<std::string> const& argDynamicSetAttrs);
	explicit LightAttributeSet(AttributeSet<T> const& attrSet);
	LightAttributeSet(LightAttributeSet<T> const& attrSet);
	T& getDefaultAtomValue();
	T const& getDefaultAtomValue() const;
	std::set<T>& getDefaultSetValue();
	std::set<T> const& getDefaultSetValue() const;
	std::map<std::string, T>* getStaticAtomAttrs() const;
	std::map<std::string, T> const& getDynamicAtomAttrs() const;
	std::map<std::string, std::set<T> >* getStaticSetAttrs() const;
	std::map<std::string, std::set<T> > const& getDynamicSetAttrs() const;
	T& getAtom(std::string const& attrName);
	const T& getAtom(std::string const& attrName) const;
	std::set<T>& getSet(std::string const& attrName);
	const std::set<T>& getSet(std::string const& attrName) const;
	bool hasAttributeOfName(std::string const& attrName) const;
	bool hasAtomOfName(std::string const& attrName) const;
	bool hasSetOfName(std::string const& attrName) const;
	bool isStatic(std::string const& attrName) const;
	bool isDynamic(std::string const& attrName) const;
	bool isAtom(std::string const& attrName) const;
	bool isSet(std::string const& attrName) const;
	void updateAttrs(std::map<T,T> const& oldToNewValues);
	void getAllAttributeNames(std::list<std::string>& attrList) const;
	void toStringStream(std::ostringstream& st) const;
	bool empty() const;
	~LightAttributeSet();
};

/** Constructor: empty, create an empty LightAttributeSet.
 * Initialize defaultAtomValue to argDefaultValue and defaultSetValue to an empty set. Initialise the set pointers to NULL.
 * @param argDefaultValue The default value for the atom attributes of type T
 */
template<typename T> inline LightAttributeSet<T>::LightAttributeSet(T argDefaultValue) : defaultAtomValue(argDefaultValue) {
	defaultSetValue = std::set<T>();
	staticAtomAttrs = NULL;
	staticSetAttrs = NULL;
}

/** Constructor: store data inside arguments depending on their type static/dynamic and atom/set.
 * The lists given as argument contain the name of the attributes, all will be initialised to the default value.
 * @param argDefaultValue The defalut value for the atom attributes of type T
 * @param argStaticAtomAttrs Pointer to a static-atom attribute
 * @param argDynamicAtomAttrs Initialise the dynamic-atom-attribute list
 * @param argStaticSetAttrs Pointer to a static-set attribute
 * @param argDynamicSetAttrs Initialise the dynamic-set-attribute list
 */
template<typename T> inline LightAttributeSet<T>::LightAttributeSet(T argDefaultValue, std::map<std::string, T>* argStaticAtomAttrs,
		std::vector<std::string> const& argDynamicAtomAttrs, std::map<std::string, std::set<T> >* argStaticSetAttrs,
		std::vector<std::string> const& argDynamicSetAttrs) : defaultAtomValue(argDefaultValue) {
	defaultSetValue = std::set<T>();
	size_t i;
	staticAtomAttrs = argStaticAtomAttrs;
	for(i=0 ; i<argDynamicAtomAttrs.size() ; ++i) dynamicAtomAttrs[argDynamicAtomAttrs[i]] = defaultAtomValue;
	staticSetAttrs = argStaticSetAttrs;
	for(i=0 ; i<argDynamicSetAttrs.size() ; ++i) dynamicSetAttrs[argDynamicSetAttrs[i]] = defaultSetValue;
}

/** Constructor from an AttributeSet.
 * It creates pointer to static attributes (set and atom) and copies the dynamic attributes.
 * @param attrSet AttributeSet to copy
 */
template<typename T> inline LightAttributeSet<T>::LightAttributeSet(AttributeSet<T> const& attrSet) {
	defaultAtomValue = attrSet.getDefaultAtomValue();
	defaultSetValue = attrSet.getDefaultSetValue();
	staticAtomAttrs = attrSet.getPointerOnStaticAtomAttrs();
	dynamicAtomAttrs = attrSet.getDynamicAtomAttrs();
	staticSetAttrs = attrSet.getPointerOnStaticSetAttrs();
	dynamicSetAttrs = attrSet.getDynamicSetAttrs();
}

/** Copy constructor.
 * @param attrSet LightAttributeSet to copy
 */
template<typename T> inline LightAttributeSet<T>::LightAttributeSet(LightAttributeSet<T> const& attrSet) {
	defaultAtomValue = attrSet.getDefaultAtomValue();
	defaultSetValue = attrSet.getDefaultSetValue();
	staticAtomAttrs = attrSet.getStaticAtomAttrs();
	dynamicAtomAttrs = attrSet.getDynamicAtomAttrs();
	staticSetAttrs = attrSet.getStaticSetAttrs();
	dynamicSetAttrs = attrSet.getDynamicSetAttrs();
}

/** Return the default value for atom attributes.
 * @return Default atom value
 */
template<typename T> inline T& LightAttributeSet<T>::getDefaultAtomValue() {
	return defaultAtomValue;
}

/** Return the default value for atom attributes (const version).
 * @return Default atom value
 */
template<typename T> inline T const& LightAttributeSet<T>::getDefaultAtomValue() const {
	return defaultAtomValue;
}

/** Return the default set for an attribute.
 * @return Default set value
 */
template<typename T> inline std::set<T>& LightAttributeSet<T>::getDefaultSetValue() {
	return defaultSetValue;
}

/** Return the default set for an attribute (const version).
 * @return DefaultSetValue
 */
template<typename T> inline std::set<T> const& LightAttributeSet<T>::getDefaultSetValue() const {
	return defaultSetValue;
}

/** Return a pointer to the set of static atom attributes and their value.
 * @return Pointer to the static atom attributes and their value
 */
template<typename T> inline std::map<std::string, T>* LightAttributeSet<T>::getStaticAtomAttrs() const {
	return staticAtomAttrs;
}

/** Return the set of dynamic atom attributes and their value.
 * @return Dynamic atom attributes and their value
 */
template<typename T> inline std::map<std::string, T> const& LightAttributeSet<T>::getDynamicAtomAttrs() const {
	return dynamicAtomAttrs;
}

/** Return a pointer to the set of static set attributes and their value.
 * @return Pointer to the static set attributes and their value
 */
template<typename T> inline std::map<std::string, std::set<T> >* LightAttributeSet<T>::getStaticSetAttrs() const {
	return staticSetAttrs;
}

/** Return the set of dynamic set attributes and their value.
 * @return Dynamic set attributes and their value
 */
template<typename T> inline std::map<std::string, std::set<T> > const& LightAttributeSet<T>::getDynamicSetAttrs() const {
	return dynamicSetAttrs;
}

/** Return a reference to the value of the atom attribute attrName.
 * @param attrName Name of the atom attribute
 * @return Reference on the value of attrName
 * @throw Throw an HATPexception if attrName is not the name of an existing attribute
 */
template<typename T> inline T& LightAttributeSet<T>::getAtom(std::string const& attrName) {
	typename std::map<std::string, T>::iterator it;
	it = staticAtomAttrs->find(attrName);
	if(it != staticAtomAttrs->end()) return it->second;
	it = dynamicAtomAttrs.find(attrName);
	if(it != dynamicAtomAttrs.end()) return it->second;
	throw HATPexception(ERROR,"trying to access to a non existing attribute "+attrName);
	return defaultAtomValue;
}

/** Return a const reference to the value of the atom attribute attrName.
 * @param attrName Name of the atom attribute
 * @return Const reference on the value of attrName
 * @throw Throw an HATPexception if attrName is not the name of an existing attribute
 */
template<typename T> inline T const& LightAttributeSet<T>::getAtom(std::string const& attrName) const {
	typename std::map<std::string, T>::const_iterator it;
	it = staticAtomAttrs->find(attrName);
	if(it != staticAtomAttrs->end()) return it->second;
	it = dynamicAtomAttrs.find(attrName);
	if(it != dynamicAtomAttrs.end()) return it->second;
	throw HATPexception(ERROR,"trying to get the value of a non existing attribute "+attrName);
	return defaultAtomValue;
}

/** Return a reference to the value of the set attribute attrName.
 * @param attrName Name of the set attribute
 * @return Reference on the value of attrName
 * @throw Throw an HATPexception if attrName is not the name of an existing attribute
 */
template<typename T>
inline std::set<T>& LightAttributeSet<T>::getSet(std::string const& attrName) {
	typename std::map<std::string, std::set<T> >::iterator it;
	it = staticSetAttrs->find(attrName);
	if(it != staticSetAttrs->end()) return it->second;
	it = dynamicSetAttrs.find(attrName);
	if(dynamicSetAttrs.find(attrName) != dynamicSetAttrs.end()) return it->second;
	throw HATPexception(ERROR,"trying to access to a non existing attribute "+attrName);
	return defaultSetValue;
}

/** Return a const reference to the value of the set attribute attrName.
 * @param attrName Name of the set attribute
 * @return Const reference on the value of attrName
 * @throw Throw an HATPexception if attrName is not the name of an existing attribute
 */
template<typename T> inline std::set<T> const& LightAttributeSet<T>::getSet(std::string const& attrName) const {
	typename std::map<std::string, std::set<T> >::const_iterator it;
	it = staticSetAttrs->find(attrName);
	if(it != staticSetAttrs->end()) return it->second;
	it = dynamicSetAttrs.find(attrName);
	if(dynamicSetAttrs.find(attrName) != dynamicSetAttrs.end()) return it->second;
	throw HATPexception(ERROR,"trying to access to a non existing attribute "+attrName);
	return defaultSetValue;
}

/** Check if an attribute with the name attrName exists (either as a static/dynamic and atom/set).
 * @param attrName The attribute name to check
 * @return True if there is an attribute named attrName
 */
template<typename T> inline bool LightAttributeSet<T>::hasAttributeOfName(std::string const& attrName) const {
	return staticAtomAttrs->find(attrName) != staticAtomAttrs->end() ||
	 dynamicAtomAttrs.find(attrName) != dynamicAtomAttrs.end() ||
	 staticSetAttrs->find(attrName) != staticSetAttrs->end() ||
	 dynamicSetAttrs.find(attrName) != dynamicSetAttrs.end();
}

/** Check if an atom attribute exists with the name attrName (either static or dynamic).
 * @param attrName The attribute name to check
 * @return True if an attribute exits
 */
template<typename T> inline bool LightAttributeSet<T>::hasAtomOfName(std::string const& attrName) const {
	return staticAtomAttrs->find(attrName) != staticAtomAttrs->end() ||
	 dynamicAtomAttrs.find(attrName) != dynamicAtomAttrs.end();
}

/** Check if a set attribute exists with the name attrName (either static or dynamic).
 * @param attrName The attribute name to check
 * @return True if an attribute exits
 */
template<typename T> inline bool LightAttributeSet<T>::hasSetOfName(std::string const& attrName) const {
	return staticSetAttrs->find(attrName) != staticSetAttrs->end() ||
	 dynamicSetAttrs.find(attrName) != dynamicSetAttrs.end();
}

/** Check if the attribute attrName is either static or not (the attribute can be an atom or set attribute).
 * @param attrName The attribute the check
 * @return True if the attribute is static
 */
template<typename T> inline bool LightAttributeSet<T>::isStatic(std::string const& attrName) const {
	return staticAtomAttrs->find(attrName) != staticAtomAttrs->end() ||
	 staticSetAttrs->find(attrName) != staticSetAttrs->end();
}

/** Check if the attribute attrName is either dynamic or not (the attribute can be an atom or set attribute).
 * @param attrName The attribute the check
 * @return True if the attribute is dynamic
 */
template<typename T> inline bool LightAttributeSet<T>::isDynamic(std::string const& attrName) const {
	return dynamicAtomAttrs.find(attrName) != dynamicAtomAttrs.end() ||
	 dynamicSetAttrs.find(attrName) != dynamicSetAttrs.end();
}

/** Check if the attribute attrName is either an atom attribute or not (the attribute can be static or dynamic).
 * @param attrName The attribute the check
 * @return True if the attribute is an atom attribute
 */
template<typename T> inline bool LightAttributeSet<T>::isAtom(std::string const& attrName) const {
	return staticAtomAttrs->find(attrName) != staticAtomAttrs->end() ||
	 dynamicAtomAttrs.find(attrName) != dynamicAtomAttrs.end();
}

/** Check if the attribute attrName is either a set attribute or not (the attribute can be static or dynamic).
 * @param attrName The attribute the check
 * @return True if the attribute is a set attribute
 */
template<typename T> inline bool LightAttributeSet<T>::isSet(std::string const& attrName) const {
	return staticSetAttrs->find(attrName) != staticSetAttrs->end() ||
	 dynamicSetAttrs.find(attrName) != dynamicSetAttrs.end();
}

/** Update the values of attributes.
 * Explores all attributes and when one has a value NewValues->first, replace this value by NewValues->second.
 * @param oldToNewValues A map where the keys are the old values and the content for a key is the new value
 */
template<typename T> inline void LightAttributeSet<T>::updateAttrs(std::map<T,T> const& oldToNewValues) {
	typename std::map<std::string, T>::iterator it1;
	typename std::map<std::string, std::set<T> >::iterator it2;
	typename std::map<T, T>::const_iterator itVal;
	typename std::set<T>::iterator itSet;
	for(itVal = oldToNewValues.begin() ; itVal != oldToNewValues.end() ; ++itVal) {
		for(it1=dynamicAtomAttrs.begin() ; it1 != dynamicAtomAttrs.end() ; ++it1)
			if(it1->second == itVal->first)
				it1->second = itVal->second;
		for(it2=dynamicSetAttrs.begin() ; it2 != dynamicSetAttrs.end() ; ++it2)
			for(itSet = (it2->second).begin() ; itSet != (it2->second).end() ; ++itSet)
				if(*itSet == itVal->first) {
					(it2->second).erase(itSet);
					(it2->second).insert(itVal->second);
				}
	}
}

/** Return the list of all attributes (static and dynamic, atom and set).
 * Complete attrList with attribute names of the LightAttributeSet.
 * @param attrList The list that is filled by the all the attributes' name
 */
template<typename T> inline void LightAttributeSet<T>::getAllAttributeNames(std::list<std::string>& attrList) const {
	typename std::map<std::string, T>::const_iterator it1;
	typename std::map<std::string, std::set<T> >::const_iterator it2;
	for(it1=staticAtomAttrs->begin() ; it1 != staticAtomAttrs->end() ; ++it1)
		attrList.push_back(it1->first);
	for(it1=dynamicAtomAttrs.begin() ; it1 != dynamicAtomAttrs.end() ; ++it1)
		attrList.push_back(it1->first);
	for(it2=staticSetAttrs->begin() ; it2 != staticSetAttrs->end() ; ++it2)
		attrList.push_back(it2->first);
	for(it2=dynamicSetAttrs.begin() ; it2 != dynamicSetAttrs.end() ; ++it2)
		attrList.push_back(it2->first);
}

/** Print all the attributes to the stream givean as argument.
 * Print the attributes, their type and their value. Like :
 * - name : value (static)
 * - name : value (dynamic)
 * - name : {value, value, ... } (static)
 * - name : {value, value, ... } (dynamic)
 *
 * @param st The stream the attributes will be printed in
 */
template<typename T> inline void LightAttributeSet<T>::toStringStream(std::ostringstream& st) const {
	typename std::map<std::string, T>::const_iterator it1;
	typename std::map<std::string, std::set<T> >::const_iterator it2;
	typename std::set<T>::const_iterator setIT;
	std::set<T> setData;
	for(it1=staticAtomAttrs->begin() ; it1 != staticAtomAttrs->end() ; ++it1)
		st << "\t\t" << it1->first + ": " << it1->second << " (static)" << std::endl;
	for(it1=dynamicAtomAttrs.begin() ; it1 != dynamicAtomAttrs.end() ; ++it1)
		st << "\t\t" << it1->first + ": " << it1->second << " (dynamic)" << std::endl;
	for(it2=staticSetAttrs->begin() ; it2 != staticSetAttrs->end() ; ++it2) {
		setData = it2->second;
		st << "\t\t" << it2->first << ": {";
		for(setIT=setData.begin() ; setIT != setData.end() ;) {
			st << *setIT;
			if(++setIT != setData.end()) st << ", ";
		}
		st << "} (static)" << std::endl;
	}
	for(it2=dynamicSetAttrs.begin() ; it2 != dynamicSetAttrs.end() ; ++it2) {
		setData = it2->second;
		st << "\t\t" << it2->first << ": {";
		for(setIT=setData.begin() ; setIT != setData.end() ;) {
			st << *setIT;
			if(++setIT != setData.end()) st << ", ";
		}
		st << "} (dynamic)" << std::endl;
	}
}

/** Specialised version of toStringStream for bool attributes.
 * See LightAttributeSet<T>::toStringStream(std::ostringstream& st)
 * @param st The stream the attributes will be printed in
 */
template<> inline void LightAttributeSet<bool>::toStringStream(std::ostringstream& st) const {
	std::map<std::string, bool>::const_iterator it1;
	std::map<std::string, std::set<bool> >::const_iterator it2;
	std::set<bool>::const_iterator setIT;
	std::set<bool> setData;
	std::string sTrue("true"), sFalse("false");
	for(it1=staticAtomAttrs->begin() ; it1 != staticAtomAttrs->end() ; ++it1) {
		st << "\t\t" << it1->first + ": ";
		if(it1->second) st << sTrue;
		else st << sFalse;
		st << " (static)" << std::endl;
	}
	for(it1=dynamicAtomAttrs.begin() ; it1 != dynamicAtomAttrs.end() ; ++it1) {
		st << "\t\t" << it1->first + ": ";
		if(it1->second) st << sTrue;
		else st << sFalse;
		st << " (dynamic)" << std::endl;
	}
	for(it2=staticSetAttrs->begin() ; it2 != staticSetAttrs->end() ; ++it2) {
		setData = it2->second;
		st << "\t\t" << it2->first << ": {";
		for(setIT=setData.begin() ; setIT != setData.end() ;) {
			if(*setIT) st << sTrue;
			else st << sFalse;
			if(++setIT != setData.end()) st << ", ";
		}
		st << "} (static)" << std::endl;
	}
	for(it2=dynamicSetAttrs.begin() ; it2 != dynamicSetAttrs.end() ; ++it2) {
		setData = it2->second;
		st << "\t\t" << it2->first << ": {";
		for(setIT=setData.begin() ; setIT != setData.end() ;) {
			if(*setIT) st << sTrue;
			else st << sFalse;
			if(++setIT != setData.end()) st << ", ";
		}
		st << "} (dynamic)" << std::endl;
	}
}

/** Check if all attribute list are empty.
 * @return True if the LightAttributeSet is empty
 */
template<typename T> inline bool LightAttributeSet<T>::empty() const{
	return staticAtomAttrs->empty() && dynamicAtomAttrs.empty() && staticSetAttrs->empty() && dynamicSetAttrs.empty();
}

/** Destructor: empty
 */
template<typename T> inline LightAttributeSet<T>::~LightAttributeSet(){}

#endif
